@@ -1,6 +1,8 @@
 # This file is part of Tryton.  The COPYRIGHT file at the top level of
 # this repository contains the full copyright notices and license terms.
 from decimal import Decimal
+from collections import defaultdict
+
 from sql import Column
 from sql.aggregate import Sum
 from sql.conditionals import Coalesce
@@ -8,11 +10,12 @@ from sql.conditionals import Coalesce
 from trytond import backend
 from trytond.model import ModelView, ModelSQL, fields, Unique
 from trytond.wizard import Wizard, StateView, StateAction, Button
-from trytond.pyson import Eval, PYSONEncoder
+from trytond.pyson import Eval, If, PYSONEncoder, PYSONDecoder
 from trytond.transaction import Transaction
 from trytond.pool import Pool
 
-__all__ = ['Account', 'OpenChartAccountStart', 'OpenChartAccount',
+__all__ = ['Account', 'AccountDistribution',
+    'OpenChartAccountStart', 'OpenChartAccount',
     'AnalyticAccountEntry', 'AnalyticMixin']
 
 
@@ -22,17 +25,21 @@ class Account(ModelSQL, ModelView):
     name = fields.Char('Name', required=True, translate=True, select=True)
     code = fields.Char('Code', select=True)
     active = fields.Boolean('Active', select=True)
-    company = fields.Many2One('company.company', 'Company')
-    currency = fields.Many2One('currency.currency', 'Currency', required=True)
+    company = fields.Many2One('company.company', 'Company', required=True)
+    currency = fields.Function(
+        fields.Many2One('currency.currency', 'Currency'),
+        'on_change_with_currency')
     currency_digits = fields.Function(fields.Integer('Currency Digits'),
         'on_change_with_currency_digits')
     type = fields.Selection([
         ('root', 'Root'),
         ('view', 'View'),
         ('normal', 'Normal'),
+        ('distribution', 'Distribution'),
         ], 'Type', required=True)
     root = fields.Many2One('analytic_account.account', 'Root', select=True,
         domain=[
+            ('company', '=', Eval('company', -1)),
             ('parent', '=', None),
             ('type', '=', 'root'),
             ],
@@ -40,15 +47,25 @@ class Account(ModelSQL, ModelView):
             'invisible': Eval('type') == 'root',
             'required': Eval('type') != 'root',
             },
-        depends=['type'])
+        depends=['company', 'type'])
     parent = fields.Many2One('analytic_account.account', 'Parent', select=True,
-        domain=[('parent', 'child_of', Eval('root'))],
+        domain=['OR',
+            ('root', '=', Eval('root', -1)),
+            ('parent', '=', None),
+            ],
         states={
             'invisible': Eval('type') == 'root',
             'required': Eval('type') != 'root',
             },
         depends=['root', 'type'])
-    childs = fields.One2Many('analytic_account.account', 'parent', 'Children')
+    childs = fields.One2Many('analytic_account.account', 'parent', 'Children',
+        states={
+            'invisible': Eval('id', -1) < 0,
+            },
+        domain=[
+            ('company', '=', Eval('company', -1)),
+            ],
+        depends=['company'])
     balance = fields.Function(fields.Numeric('Balance',
         digits=(16, Eval('currency_digits', 1)), depends=['currency_digits']),
         'get_balance')
@@ -71,12 +88,39 @@ class Account(ModelSQL, ModelView):
     mandatory = fields.Boolean('Mandatory', states={
             'invisible': Eval('type') != 'root',
             },
+        depends=['type'],
+        help="Make this account mandatory when filling documents")
+    distributions = fields.One2Many(
+        'analytic_account.account.distribution', 'parent',
+        "Distributions",
+        states={
+            'invisible': Eval('type') != 'distribution',
+            'required': Eval('type') == 'distribution',
+            },
         depends=['type'])
+    distribution_parents = fields.Many2Many(
+        'analytic_account.account.distribution', 'account', 'parent',
+        "Distribution Parents", readonly=True)
 
     @classmethod
     def __setup__(cls):
         super(Account, cls).__setup__()
         cls._order.insert(0, ('code', 'ASC'))
+        cls._order.insert(1, ('name', 'ASC'))
+        cls._error_messages.update({
+                'invalid_distribution': (
+                    'The distribution sum of account "%(account)s" '
+                    'is not 100%.'),
+                })
+
+    @classmethod
+    def __register__(cls, module_name):
+        TableHandler = backend.get('TableHandler')
+        super(Account, cls).__register__(module_name)
+        table = TableHandler(cls, module_name)
+
+        # Migration from 4.0: remove currency
+        table.not_null_action('currency', action='remove')
 
     @staticmethod
     def default_active():
@@ -86,13 +130,6 @@ class Account(ModelSQL, ModelView):
     def default_company():
         return Transaction().context.get('company')
 
-    @staticmethod
-    def default_currency():
-        Company = Pool().get('company.company')
-        if Transaction().context.get('company'):
-            company = Company(Transaction().context['company'])
-            return company.currency.id
-
     @staticmethod
     def default_type():
         return 'normal'
@@ -113,28 +150,49 @@ class Account(ModelSQL, ModelView):
     def validate(cls, accounts):
         super(Account, cls).validate(accounts)
         cls.check_recursion(accounts)
+        cls.check_recursion(accounts, parent='distribution_parents')
+        for account in accounts:
+            account.check_distribution()
 
-    @fields.depends('currency')
+    def check_distribution(self):
+        if self.type != 'distribution':
+            return
+        if sum((d.ratio for d in self.distributions)) != 1:
+            self.raise_user_error('invalid_distribution', {
+                    'account': self.rec_name,
+                    })
+
+    @fields.depends('company')
+    def on_change_with_currency(self, name=None):
+        if self.company:
+            return self.company.currency.id
+
+    @fields.depends('company')
     def on_change_with_currency_digits(self, name=None):
-        if self.currency:
-            return self.currency.digits
+        if self.company:
+            return self.company.currency.digits
         return 2
 
+    @fields.depends('parent', 'type',
+        '_parent_parent.root', '_parent_parent.type')
+    def on_change_parent(self):
+        if self.parent and self.type != 'root':
+            if self.parent.type == 'root':
+                self.root = self.parent
+            else:
+                self.root = self.parent.root
+        else:
+            self.root = None
+
     @classmethod
     def get_balance(cls, accounts, name):
-        res = {}
         pool = Pool()
         Line = pool.get('analytic_account.line')
         MoveLine = pool.get('account.move.line')
-        Account = pool.get('account.account')
-        Company = pool.get('company.company')
-        Currency = pool.get('currency.currency')
         cursor = Transaction().connection.cursor()
         table = cls.__table__()
         line = Line.__table__()
         move_line = MoveLine.__table__()
-        a_account = Account.__table__()
-        company = Company.__table__()
 
         ids = [a.id for a in accounts]
         childs = cls.search([('parent', 'child_of', ids)])
@@ -150,67 +208,42 @@ class Account(ModelSQL, ModelView):
                 condition=table.id == line.account
                 ).join(move_line, 'LEFT',
                 condition=move_line.id == line.move_line
-                ).join(a_account, 'LEFT',
-                condition=a_account.id == move_line.account
-                ).join(company, 'LEFT',
-                condition=company.id == a_account.company
                 ).select(table.id,
                 Sum(Coalesce(line.debit, 0) - Coalesce(line.credit, 0)),
-                company.currency,
                 where=(table.type != 'view')
                 & table.id.in_(all_ids)
                 & (table.active == True) & line_query,
-                group_by=(table.id, company.currency)))
-        account_sum = {}
-        id2currency = {}
-        for account_id, sum, currency_id in cursor.fetchall():
+                group_by=table.id))
+        account_sum = defaultdict(Decimal)
+        for account_id, value in cursor.fetchall():
             account_sum.setdefault(account_id, Decimal('0.0'))
             # SQLite uses float for SUM
-            if not isinstance(sum, Decimal):
-                sum = Decimal(str(sum))
-            if currency_id != id2account[account_id].currency.id:
-                currency = None
-                if currency_id in id2currency:
-                    currency = id2currency[currency_id]
-                else:
-                    currency = Currency(currency_id)
-                    id2currency[currency.id] = currency
-                account_sum[account_id] += Currency.compute(currency, sum,
-                        id2account[account_id].currency, round=True)
-            else:
-                account_sum[account_id] += \
-                    id2account[account_id].currency.round(sum)
+            if not isinstance(value, Decimal):
+                value = Decimal(str(value))
+            account_sum[account_id] += value
 
-        for account_id in ids:
-            res.setdefault(account_id, Decimal('0.0'))
+        balances = {}
+        for account in accounts:
+            balance = Decimal()
             childs = cls.search([
-                    ('parent', 'child_of', [account_id]),
+                    ('parent', 'child_of', [account.id]),
                     ])
-            to_currency = id2account[account_id].currency
             for child in childs:
-                from_currency = id2account[child.id].currency
-                res[account_id] += Currency.compute(from_currency,
-                        account_sum.get(child.id, Decimal('0.0')), to_currency,
-                        round=True)
-            res[account_id] = to_currency.round(res[account_id])
-            if id2account[account_id].display_balance == 'credit-debit':
-                res[account_id] = - res[account_id]
-        return res
+                balance += account_sum[child.id]
+            if account.display_balance == 'credit-debit' and balance:
+                balance *= -1
+            balances[account.id] = account.currency.round(balance)
+        return balances
 
     @classmethod
     def get_credit_debit(cls, accounts, names):
         pool = Pool()
         Line = pool.get('analytic_account.line')
         MoveLine = pool.get('account.move.line')
-        Account = pool.get('account.account')
-        Company = pool.get('company.company')
-        Currency = pool.get('currency.currency')
         cursor = Transaction().connection.cursor()
         table = cls.__table__()
         line = Line.__table__()
         move_line = MoveLine.__table__()
-        a_account = Account.__table__()
-        company = Company.__table__()
 
         result = {}
         ids = [a.id for a in accounts]
@@ -224,43 +257,31 @@ class Account(ModelSQL, ModelView):
             id2account[account.id] = account
 
         line_query = Line.query_get(line)
-        columns = [table.id, company.currency]
+        columns = [table.id]
         for name in names:
             columns.append(Sum(Coalesce(Column(line, name), 0)))
         cursor.execute(*table.join(line, 'LEFT',
                 condition=table.id == line.account
                 ).join(move_line, 'LEFT',
                 condition=move_line.id == line.move_line
-                ).join(a_account, 'LEFT',
-                condition=a_account.id == move_line.account
-                ).join(company, 'LEFT',
-                condition=company.id == a_account.company
                 ).select(*columns,
                 where=(table.type != 'view')
                 & table.id.in_(ids)
                 & (table.active == True) & line_query,
-                group_by=(table.id, company.currency)))
+                group_by=table.id))
 
-        id2currency = {}
         for row in cursor.fetchall():
-            account = id2account[row[0]]
-            currency_id = row[1]
-            for i, name in enumerate(names, 2):
+            account_id = row[0]
+            for i, name in enumerate(names, 1):
+                value = row[i]
                 # SQLite uses float for SUM
-                sum = row[i]
-                if not isinstance(sum, Decimal):
-                    sum = Decimal(str(sum))
-                if currency_id != account.currency.id:
-                    currency = None
-                    if currency_id in id2currency:
-                        currency = id2currency[currency_id]
-                    else:
-                        currency = Currency(currency_id)
-                        id2currency[currency.id] = currency
-                    result[name][account.id] += Currency.compute(currency, sum,
-                            account.currency, round=True)
-                else:
-                    result[name][account.id] += account.currency.round(sum)
+                if not isinstance(value, Decimal):
+                    value = Decimal(str(value))
+                result[name][account_id] += value
+        for account in accounts:
+            for name in names:
+                result[name][account.id] = account.currency.round(
+                    result[name][account.id])
         return result
 
     def get_rec_name(self, name):
@@ -277,6 +298,30 @@ class Account(ModelSQL, ModelView):
         else:
             return [(cls._rec_name,) + tuple(clause[1:])]
 
+    def distribute(self, amount):
+        "Return a list of (account, amount) distribution"
+        assert self.type in {'normal', 'distribution'}
+        if self.type == 'normal':
+            return [(self, amount)]
+        else:
+            result = []
+            remainder = amount
+            for distribution in self.distributions:
+                account = distribution.account
+                ratio = distribution.ratio
+                current_amount = self.currency.round(amount * ratio)
+                remainder -= current_amount
+                result.extend(account.distribute(current_amount))
+            if remainder:
+                i = 0
+                while remainder:
+                    account, amount = result[i]
+                    rounding = self.currency.rounding.copy_sign(remainder)
+                    result[i] = (account, amount - rounding)
+                    remainder -= rounding
+                    i = (i + 1) % len(result)
+            return result
+
 
 class OpenChartAccountStart(ModelView):
     'Open Chart of Accounts'
@@ -306,12 +351,50 @@ class OpenChartAccount(Wizard):
         return 'end'
 
 
+class AccountDistribution(ModelView, ModelSQL):
+    "Analytic Account Distribution"
+    __name__ = 'analytic_account.account.distribution'
+    parent = fields.Many2One(
+        'analytic_account.account', "Parent", required=True, select=True)
+    root = fields.Function(
+        fields.Many2One('analytic_account.account', "Root"),
+        'on_change_with_root')
+    account = fields.Many2One(
+        'analytic_account.account', "Account", required=True,
+        domain=[
+            ('root', '=', Eval('root', -1)),
+            ],
+        depends=['root'])
+    ratio = fields.Numeric("Ratio", required=True,
+        domain=[
+            ('ratio', '>=', 0),
+            ('ratio', '<=', 1),
+            ])
+
+    @classmethod
+    def __setup__(cls):
+        super(AccountDistribution, cls).__setup__()
+        cls._order.insert(0, ('ratio', 'DESC'))
+
+    @fields.depends('parent', '_parent_parent.root')
+    def on_change_with_root(self, name=None):
+        if self.parent:
+            return self.parent.root.id
+
+
 class AnalyticAccountEntry(ModelView, ModelSQL):
     'Analytic Account Entry'
     __name__ = 'analytic.account.entry'
     origin = fields.Reference('Origin', selection='get_origin', select=True)
     root = fields.Many2One('analytic_account.account', 'Root Analytic',
-        domain=[('type', '=', 'root')])
+        domain=[
+            If(~Eval('company'),
+                # No constraint if the origin is not set
+                (),
+                ('company', '=', Eval('company', -1))),
+            ('type', '=', 'root'),
+            ],
+        depends=['company'])
     account = fields.Many2One('analytic_account.account', 'Account',
         ondelete='RESTRICT',
         states={
@@ -319,11 +402,13 @@ class AnalyticAccountEntry(ModelView, ModelSQL):
             },
         domain=[
             ('root', '=', Eval('root')),
-            ('type', '=', 'normal'),
+            ('type', 'in', ['normal', 'distribution']),
             ],
-        depends=['root', 'required'])
+        depends=['root', 'required', 'company'])
     required = fields.Function(fields.Boolean('Required'),
         'on_change_with_required')
+    company = fields.Function(fields.Many2One('company.company', 'Company'),
+        'on_change_with_company', searcher='search_company')
 
     @classmethod
     def __register__(cls, module_name):
@@ -386,8 +471,31 @@ class AnalyticAccountEntry(ModelView, ModelSQL):
             return self.root.mandatory
         return False
 
+    def on_change_with_company(self, name=None):
+        return None
 
-class AnalyticMixin(ModelSQL):
+    @classmethod
+    def search_company(cls, name, clause):
+        return []
+
+    def get_analytic_lines(self, line, date):
+        "Yield analytic lines for the accounting line and the date"
+        pool = Pool()
+        AnalyticLine = pool.get('analytic_account.line')
+
+        if not self.account:
+            return
+        amount = line.debit or line.credit
+        for account, amount in self.account.distribute(amount):
+            analytic_line = AnalyticLine()
+            analytic_line.debit = amount if line.debit else Decimal(0)
+            analytic_line.credit = amount if line.credit else Decimal(0)
+            analytic_line.account = account
+            analytic_line.date = date
+            yield analytic_line
+
+
+class AnalyticMixin(object):
 
     analytic_accounts = fields.One2Many('analytic.account.entry', 'origin',
         'Analytic Accounts',
@@ -427,13 +535,21 @@ class AnalyticMixin(ModelSQL):
                         where=entry.selection == selection_id))
             handler.drop_column('analytic_accounts')
 
-    @staticmethod
-    def default_analytic_accounts():
+    @classmethod
+    def analytic_accounts_domain(cls):
+        context = Transaction().context.copy()
+        context['context'] = context
+        return PYSONDecoder(context).decode(
+            PYSONEncoder().encode(cls.analytic_accounts.domain))
+
+    @classmethod
+    def default_analytic_accounts(cls):
         pool = Pool()
         AnalyticAccount = pool.get('analytic_account.account')
 
         accounts = []
-        root_accounts = AnalyticAccount.search([
+        root_accounts = AnalyticAccount.search(
+            cls.analytic_accounts_domain() + [
                 ('parent', '=', None),
                 ])
         for account in root_accounts:
@@ -443,11 +559,14 @@ class AnalyticMixin(ModelSQL):
                     })
         return accounts
 
-    @staticmethod
-    def default_analytic_accounts_size():
+    @classmethod
+    def default_analytic_accounts_size(cls):
         pool = Pool()
         AnalyticAccount = pool.get('analytic_account.account')
-        return len(AnalyticAccount.search([('type', '=', 'root')]))
+        return len(AnalyticAccount.search(
+            cls.analytic_accounts_domain() + [
+                    ('type', '=', 'root'),
+                    ]))
 
     @classmethod
     def get_analytic_accounts_size(cls, records, name):
@@ -464,12 +583,17 @@ class AnalyticMixin(ModelSQL):
         "Check that all mandatory root entries are defined in entries"
         pool = Pool()
         Account = pool.get('analytic_account.account')
-        mandatory_roots = {a for a in Account.search([
+        all_mandatory_roots = {a for a in Account.search([
                 ('type', '=', 'root'),
                 ('mandatory', '=', True),
                 ])}
         for analytic in analytics:
             analytic_roots = {e.root for e in analytic.analytic_accounts}
+            companies = {e.company for e in analytic.analytic_accounts}
+            mandatory_roots = set()
+            for mandatory in all_mandatory_roots:
+                if mandatory.company in companies:
+                    mandatory_roots.add(mandatory)
             if not mandatory_roots <= analytic_roots:
                 cls.raise_user_error('root_account', {
                         'name': analytic.rec_name,
